import { Timing, UUID } from 'necessary-utils-core';

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */

function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

class QueueSettings {
  constructor() {
    this._queueRestartTime = 100;
    this.QueueCoolDownTime = 0;
  }

  get QueueRestartTime() {
    return this._queueRestartTime;
  }

  set QueueRestartTime(value) {
    if (value < 100) this._queueRestartTime = 100;else this._queueRestartTime = value;
  }

}

/**
 * # Simple Queue System
 * ### Methods
 * - PushMessage
 * - Comsume
 * - ACK
 * - Clear
 * @example
 * `Define your queue`
 * new SimpleQueue<MessageObject>();
 * @example
 * `Publish a message`
 * new SimpleQueue<MessageObject>().Push(MessageObject);
 * @example
 * `Full example`
 * var queue = new SimpleQueue<MessageObject>();
 * queue.Consume(async (message: MessageObject) => {
 *  do something...
 * });
 * queue.Push(MessageObject);
 */

class BasicQueue {
  constructor(settings = null) {
    this._messages = [];
    this.Handler = null;
    this.AutoACK = false;
    this._queueStarted = false;
    this._currentMessage = null;
    this.Intercept = {
      WhenZeroMessage: () => {
        if (this.MessageCount === 0) throw new Error("Zero message count. Process was intercepted.");
        return this.Intercept;
      },
      WhenAnyMessage: () => {
        if (this.MessageCount > 0) throw new Error("Queue contains message. Process was intercepted.");
        return this.Intercept;
      },
      When: handler => __awaiter(this, void 0, void 0, function* () {
        let currentMessage = Object.assign({}, this._currentMessage);
        const result = yield handler(this.MessageCount, currentMessage);
        if (result !== true) throw new Error("Custom handler returns false. Process was intercepted.");
        return this.Intercept;
      }),
      WhenQueueStarted: () => {
        if (this._queueStarted) throw new Error("Queue processing right now. Process was intercepted.");
        return this.Intercept;
      },
      WhenQueueStopped: () => {
        if (!this._queueStarted) throw new Error("Queue stopped. Process was intercepted.");
        return this.Intercept;
      }
    };
    if (settings === null) this.Settings = new QueueSettings();else this.Settings = settings;
  }

  Ack() {
    if (this._currentMessage === null || this.Handler == null) return;

    const i = this._messages.indexOf(this._currentMessage);

    if (i !== -1) this._messages.splice(i, 1);
    this._currentMessage = null;
    this._queueStarted = this._messages.length > 0;
  }

  PushMessage(message) {
    this._messages.push(message);

    this.start();
  }

  Consume(handler, autoAck) {
    this.Handler = handler;
    this.AutoACK = autoAck;
  }

  get MessageCount() {
    return this._messages.length;
  }

  Clear() {
    let removedMessages = this._messages.filter(x => !x.UnAck);

    this._messages = this._messages.filter(x => !removedMessages.includes(x));
    return removedMessages;
  }

  Stop() {
    this._queueStarted = false;
  }

  start() {
    return __awaiter(this, void 0, void 0, function* () {
      if (this._queueStarted) return;
      this._queueStarted = true;

      while (this._queueStarted) {
        this._currentMessage = this._messages[0];

        if (this.Handler === null || this._currentMessage.UnAck === true) {
          yield Timing.Sleep(this.Settings.QueueRestartTime);
          continue;
        }

        yield Timing.Sleep(this.Settings.QueueCoolDownTime);

        try {
          this._currentMessage.UnAck = true;
          yield this.Handler(this._currentMessage);
        } catch (error) {
          console.error("Consumer Error Ocurred. See the details; \n", error);
        }

        if (this.AutoACK) this.Ack();
      }
    });
  }

}

class BasicQueueMessage {
  constructor() {
    this.UnAck = false;
    this.Id = UUID.V4();
    this.Date = new Date().getTime();
  }

}

export { BasicQueue, BasicQueueMessage, QueueSettings };
